/*****************************************************
 * Routines to get a move, validate a move,
 * apply a move to the board,
 * and detect the end of the game.
 *
 * Kenrick Mock 8/11/03
 *****************************************************/
#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include "globals.h"
#include "move.h"

/*****************************************************
 * IsValidPlace
 *
 * Returns 1 if the move placing is a valid one, 
 * 0 otherwise.
 *****************************************************/

int IsValidPlace(char board[XDIM][YDIM], int x, int y)
{
 /* Piece off the board */
 if ((x<1) || (x>16) || (y<1) || (y>14)) return 0;

 /* Not empty */
 if (board[x][y]!=EMPTY) return 0;

 /* Valid */
 return 1;
}

/*****************************************************
 * GetInitialPlacing
 *
 * Input from the keyboard coordinates to initially
 * place each piece.
 *****************************************************/
void GetInitialPlacing(char board[XDIM][YDIM], piece_info pieces[])
{
 int i,x,y;
 char s[10];

 for (i=1; i<6; i++) {
	 ShowBoard(board,pieces);
RedRetry:
	 printf("Red player, enter coordinates for piece %d (separate by comma):\n",i);
	 scanf("%d,%d",&x,&y);
	 if (IsValidPlace(board,x,y)) {
		 board[x][y]=i-1;
		 pieces[i-1].curx=x;
		 pieces[i-1].cury=y;
	 }
	 else {
		 goto RedRetry;
	 }
	 ShowBoard(board,pieces);
GreenRetry:
	 printf("Green player, enter coordinates for piece %d (separate by comma):\n",i);
	 scanf("%d,%d",&x,&y);
	 if (IsValidPlace(board,x,y)) {
		 board[x][y]=i+4;
		 pieces[i-1].curx=x;
		 pieces[i-1].cury=y;
	 }
	 else {
		 goto GreenRetry;
	 }
 }
 fgets(s,10,stdin);		// Eat extra newline left in buffer by scanf
}

int CountCommas(char *s)
{
	int i,cnt=0;
	for (i=0; i<strlen(s); i++) {
		if (s[i]==',') cnt++;
	}
	return cnt;
}

/*****************************************************
 * ParseMove
 * 
 * Given a move as a string, parse it into the
 * move struct that is passed in.
 *
 * 9/25/03 fixed potential buffer overrun bug
 *****************************************************/
void ParseMove(char *s, move_struct *mptr)
{
  char c;
  char rest[2000],newrest[2000];
  int i,numjumps=0,numpivots;

  mptr->numjumps=0;
  mptr->type = 'X';
  if (s[0]=='W') {
	  mptr->type = 'W';
	  sscanf(s,"%c %d,%d %d,%d",&c,&(mptr->oldx),&(mptr->oldy)
			  ,&(mptr->newx),&(mptr->newy));
  }
  else if (s[0]=='J') {
	  mptr->type = 'J';
	  numpivots = CountCommas(s)-1;
	  /* %[^\n] will read all characters up to the newline */
	  sscanf(s,"%c %d,%d %[^\n]",&c,&(mptr->oldx),&(mptr->oldy),rest);
	  while ((numjumps<numpivots) && (numjumps < 20))
	  {
		 sscanf(rest,"%d,%d%[^\n]",&(mptr->jumpx[numjumps]),
			&(mptr->jumpy[numjumps]), newrest);
		 numjumps++;
		 if (numjumps < numpivots) strncpy(rest,newrest,2000);
	  }
	  mptr->numjumps = numjumps;
  }
}

/*****************************************************
 * Validate Jump
 *
 * Returns 1 if the jump is a valid one, 0 otherwise.
 *****************************************************/
int ValidJump(char board[XDIM][YDIM], int srcx, int srcy,
		int pivotx, int pivoty, int *newx, int *newy)
{
  int dx=0,dy=0;
  int i;
  int landx, landy;
  int curx,cury;

  /* printf("srcx=%d srcy=%d pivotx=%d,%d\n",srcx,srcy,pivotx,pivoty); */
  if ((pivotx < 1) || (pivotx > 16)) return 0;
  if ((pivoty < 1) || (pivoty > 14)) return 0;
  if ((pivotx == srcx) && (pivoty == srcy)) return 0;
  if ((board[pivotx][pivoty]==EMPTY) || (board[pivotx][pivoty]==WATER)) 
  	return 0;

  /* vertical jump */
  if (srcx == pivotx) {
	  *newx = srcx;
	  /* Up */
	  if (srcy > pivoty) {
		  landy = pivoty - (srcy - pivoty);
		  if (landy<1) return 0;
		  for (i=srcy-1; i>pivoty; i--) {
			if (board[srcx][i]!=EMPTY) return 0;
		  }
		  for (i=pivoty-1; i>=landy; i--) {
			if (board[srcx][i]!=EMPTY) return 0;
		  }
	  	  *newy = landy;
		  return 1;
	  }
	  /* Down */
	  else {
		  landy = pivoty+ (pivoty - srcy);
		  if (landy>14) return 0;
		  for (i=srcy+1; i<pivoty; i++) {
			if (board[srcx][i]!=EMPTY) return 0;
		  }
		  for (i=pivoty+1; i<=landy; i++) {
			if (board[srcx][i]!=EMPTY) return 0;
		  }
	  	  *newy = landy;
		  return 1;
	  }
  }
  /* Horizontal jump */
  else if (srcy == pivoty)
  {
	  *newy = srcy;
	  /* Left */
	  if (srcx > pivotx) {
		  landx = pivotx - (srcx - pivotx);
		  if (landx<1) return 0;
		  for (i=srcx-1; i>pivotx; i--) {
			if (board[i][srcy]!=EMPTY) return 0;
		  }
		  for (i=pivotx-1; i>=landx; i--) {
			if (board[i][srcy]!=EMPTY) return 0;
		  }
	  	  *newx = landx;
		  return 1;
	  }
	  /* Right */
	  else {
		  landx = pivotx + (pivotx - srcx);
		  if (landx>16) return 0;
		  for (i=srcx+1; i<pivotx; i++) {
			if (board[i][srcy]!=EMPTY) return 0;
		  }
		  for (i=pivotx+1; i<=landx; i++) {
			if (board[i][srcy]!=EMPTY) return 0;
		  }
	  	  *newx = landx;
		  return 1;
	  }
  }
  /* diagonal jump */
  else if (abs(srcx-pivotx)==abs(srcy-pivoty)) 
  {
	if (srcx > pivotx) {
		landx = pivotx - (srcx - pivotx);
		dx = -1;
	}
	else {
		landx = pivotx + (pivotx - srcx);
		dx = 1;
	}
	if (srcy > pivoty) {
		landy = pivoty - (srcy - pivoty);
		dy = -1;
	}
	else {
		landy = pivoty + (pivoty - srcy);
		dy = 1;
	}
	if ((landy<1) || (landy>14)) return 0;
	if ((landx<1) || (landx>16)) return 0;
	if (board[landx][landy]!=EMPTY) return 0;
	curx = srcx + dx;
	cury = srcy + dy;
	while ((curx!=pivotx) && (cury!=pivoty)) {
		if (board[curx][cury]!=EMPTY) return 0;
		curx+=dx;
		cury+=dy;
	}
	curx+=dx;
	cury+=dy;
	while ((curx!=landx) && (cury!=landy)) {
		if (board[curx][cury]!=EMPTY) return 0;
		curx+=dx;
		cury+=dy;
	}
	*newx = landx;
	*newy = landy;
	return 1;
  }
  return 0;
}

/*****************************************************
 * IsValid
 *
 * Returns 1 if the move is a valid one, 0 otherwise.
 * whoseturn = 'r' or 'g' to indicate the current player.
 *****************************************************/
int IsValid(char board[XDIM][YDIM], piece_info pieces[], char whoseturn,
	    char *strMove)
{
  move_struct move;
  char c;
  int tmp;
  int i,landx,landy;
  
  ParseMove(strMove, &move);
  if (move.type =='X') return 0;
  c = board[move.oldx][move.oldy];
  if (move.type=='W') {
	  if (whoseturn=='r') {
		  if ((c>=0) && (c<=4)) {
			  if ((move.newx < 1) || (move.newx > 16)) return 0;
			  if ((move.newy < 1) || (move.newy > 14)) return 0;
			  if (board[move.newx][move.newy]!=EMPTY) return 0;
			  tmp = move.newx - move.oldx;
			  if (tmp < 0) tmp *= -1;
			  if (tmp > 1) return 0;
			  tmp = move.newy - move.oldy;
			  if (tmp < 0) tmp *= -1;
			  if (tmp > 1) return 0;
			  return 1;   /* Valid walk move */
		  }
		  else {
			  return 0;
		  }
	  }
	  else {
		  if ((c>=5) && (c<=9)) {
			  if ((move.newx < 1) || (move.newx > 16)) return 0;
			  if ((move.newy < 1) || (move.newy > 14)) return 0;
			  if (board[move.newx][move.newy]!=EMPTY) return 0;
			  tmp = move.newx - move.oldx;
			  if (tmp < 0) tmp *= -1;
			  if (tmp > 1) return 0;
			  tmp = move.newy - move.oldy;
			  if (tmp < 0) tmp *= -1;
			  if (tmp > 1) return 0;
			  return 1;   /* Valid walk move */
		  }
		  else {
			  return 0;
		  }
	  }
  }
  else {
	  /* Jumping Move */
	  if (whoseturn=='r') {
		  if ((c<0) || (c>4)) return 0;
	  }
	  else {
		  if ((c<5) || (c>9)) return 0;
	  }
	  if (move.numjumps < 1) return 0;
	  if (c==EMPTY || c==WATER) return 0;
	  landx = move.oldx;
	  landy = move.oldy;
	  printf("numjumps=%d\n",move.numjumps);
	  for (i=0; i<move.numjumps; i++) {
		  tmp= ValidJump(board,landx,landy,move.jumpx[i],move.jumpy[i],
				  &landx,&landy);
		  if (tmp==0) return 0;
	  }
	  return 1;
  }
}

/*****************************************************
 * Returns 1 if segment intersects start/finish line, 
 * 0 otherwise.
 *****************************************************/
int CrossLine(int x1, int y1, int x2, int y2)
{
	if ((x1>8) && (x2<=8)) {
		/* Entirely Below Billabong */
		if ((y1>8) && (y2>8)) return 1;
		/* Diagonally, Ended above billabong */
		else if ((y1>10) && (x1<13)) return 1;
		/* Diagonally, below billabong */
		else if ((y2>10) && (x2>4)) return 1;
	}
	return 0;
}


/*****************************************************
 * Returns 1 if segment intersects top line, 
 * 0 otherwise.
 *****************************************************/
int CrossTop(int x1, int y1, int x2, int y2)
{
	if ((x1<=8) && (x2>8)) {
		/* Entirely Above Billabong */
		if ((y1<7) && (y2<7)) return 1;
		/* Diagonally, Ended below billabong */
		else if ((y1<5) && (x1>4)) return 1;
		/* Diagonally, below billabong */
		else if ((y2<5) && (x2<13)) return 1;
	}
	return 0;
}


/*****************************************************
 * Returns 1 if path has gone around the
 * billabong, 0 otherwise.
 * Uses the point in a polygon algorithm, using point 8,8.
 * Algorithm from http://www.alienryderflex.com/polygon/
 *****************************************************/
int PointInPolygon(piece_info p, float srcX, float srcY)
{
 int i,j=0;
 int oddNodes=0;
 float pYi, pXi, pYj, pXj;
 int crosstop=0;

 for (i=0; i<p.nummoves; i++) {
	 j++;
	 if (j==p.nummoves) j=0;
	 if (CrossTop(p.x[i],p.y[i],p.x[j],p.y[j])) crosstop=1;
	 pYi = (float) p.y[i];
	 pXi = (float) p.x[i];
	 pYj = (float) p.y[j];
	 pXj = (float) p.x[j];
	 if (pYi<srcY && pYj>=srcY || pYj<srcY && pYi>=srcY) {
		 if (pXi + (srcY-pYi)/(pYj-pYi)*(pXj-pXi)<srcX) {
			 if (oddNodes==0) oddNodes=1;
			 else oddNodes=0;
		 }
	 }
 }
 return oddNodes*crosstop;
}

int SurroundsBillabong(piece_info p) 
{
 if (p.nummoves < 3) return 0;
 if (PointInPolygon(p,6.1,7)) return 1;
 
 return 0;
}


/*****************************************************
 * Add a move to the move history of a piece
 * Returns 0 if full
 *****************************************************/
int AddMove(piece_info *p, int newx, int newy)
{
  if (p->nummoves == MAXMOVES) return 0;
  p->x[p->nummoves] = newx;
  p->y[p->nummoves] = newy;
  p->nummoves = p->nummoves + 1;
  return 1;
}

/*****************************************************
 * ApplyMove
 *
 * Takes a valid move and applies it to the board.
 * This moves the actual piece on the board and
 * removes any pieces that may have crossed the
 * finish line.
 *****************************************************/
void ApplyMove(char board[XDIM][YDIM], piece_info pieces[], char *strMove)
{
  move_struct move;
  char c;
  int i,landx,landy,tmp,srcx,srcy;
  
  ParseMove(strMove, &move);
  c = board[move.oldx][move.oldy];
  if (move.type=='W') {
	  board[move.oldx][move.oldy]=EMPTY;
	  board[move.newx][move.newy]=c;
	  pieces[c].curx = move.newx;
	  pieces[c].cury = move.newy;
	  landx=move.newx;
	  landy=move.newy;
	  /* Check if we crossed the start line */
	  if (pieces[c].status==PRE_START) {
		  if ((move.oldx==9) && (move.newx==8) && (move.oldy>8)) {
			  pieces[c].status = POST_START;
			  /* Start remembering moves for polygon */
			  tmp=AddMove(&(pieces[c]),move.oldx,move.oldy);
			  tmp=AddMove(&(pieces[c]),move.newx,move.newy);
			  if (tmp==0) { 
				  fprintf(stderr,"MOVE LIST FULL.\n"); 
				  exit(0); 
			  }
		  }
	  }
	  else if (pieces[c].status==POST_START) {
		  tmp=AddMove(&(pieces[c]),move.newx,move.newy);
		  if (tmp==0) { 
			  fprintf(stderr,"MOVE LIST FULL.\n"); 
			  exit(0); 
		  }
		  /* See if we crossed the finish */
		  if ((move.oldx==9) && (move.newx==8) && (move.oldy>8)) {
			if (SurroundsBillabong(pieces[c])) {
				printf("Crossed finish!\n");
				pieces[c].status=BILLABONG;
				pieces[c].curx = -1;
				pieces[c].cury = -1;
		  		board[move.newx][move.newy]=EMPTY;
			}
		  }
	  }
  }
  else 
  /* Jumping Move */
  {
	  landx = move.oldx;
	  landy = move.oldy;
	  for (i=0; i<move.numjumps; i++) {
		  srcx = landx;
		  srcy = landy;
		  tmp= ValidJump(board,landx,landy,move.jumpx[i],move.jumpy[i],
				  &landx,&landy);
		  if (tmp==0) { 
			  fprintf(stderr,
				"Move error...i=%d, %d %d : %d %d\n",i,
				move.jumpx[i],move.jumpy[i],
				landx,landy); 
			  exit(0); 
		  }
		  /* Check if we crossed the start line */
		  if (pieces[c].status==PRE_START) {
			  if (CrossLine(srcx,srcy,landx,landy)==1) 
			  {
				  pieces[c].status = POST_START;
				  /* Start remembering moves for polygon */
				  tmp=AddMove(&(pieces[c]),srcx,srcy);
				  tmp=AddMove(&(pieces[c]),landx,landy);
				  if (tmp==0) { 
					  fprintf(stderr,"MOVE LIST FULL.\n"); 
					  exit(0); 
				  }
			  }
		  }
		  else if (pieces[c].status==POST_START) {
			  tmp=AddMove(&(pieces[c]),landx,landy);
			  if (tmp==0) { 
				  fprintf(stderr,"MOVE LIST FULL.\n"); 
				  exit(0); 
			  }
			  /* See if we passed the finish line */
			  if ((CrossLine(srcx,srcy,landx,landy)==1)  &&
			      (SurroundsBillabong(pieces[c]))) {
				  pieces[c].status = BILLABONG;
				  printf("Crossed finish!\n");
	  		  	  board[move.oldx][move.oldy]=EMPTY;
				  return;
			  }
		  }
	  }
	  if (pieces[c].status != BILLABONG) {
		  board[landx][landy]=c;
		  pieces[c].curx = landx;
		  pieces[c].cury = landy;
		  board[move.oldx][move.oldy]=EMPTY;
	  }
  }
}

/*****************************************************
 * GameOver
 *
 * Returns 1 if red won, 2 if green won, 0 otherwise.
 * Also prints out the winner if verbose is 1.  
 *****************************************************/
int GameOver(piece_info pieces[],int verbose)
{
 int i;
 int num;
 
 for (i=0,num=0; (i<5) && (num==0); i++) 
 {
	 if (pieces[i].status!=BILLABONG) num++;
 }
 if (num==0) {
	 if (verbose) {
		 printf("\n**** GAME OVER ****\nRed has won!\n");
	 }
	 return 1;
 }
 for (i=5,num=0; (i<10) && (num==0); i++) 
 {
	 if (pieces[i].status!=BILLABONG) num++;
 }
 if (num==0) {
	 if (verbose) {
		 printf("\n**** GAME OVER ****\nGreen has won!\n");
	 }
	 return 2;
 }
 return 0;
}



